{# Question Presention #}
{{question_text}}
{% if use_code %}
    {% for option in question_options %}
        {{ loop.index0 }}: {{option}}
    {% endfor %}
{% else %}
    {% for option in question_options %}
        {{ option }}
    {% endfor %}    
{% endif %}

{# Restrictions #}
{% if min_selections != None and max_selections != None and min_selections == max_selections %}
    You must select exactly {{min_selections}} options.
{% elif min_selections != None and max_selections != None %}
    Minimum number of options that must be selected: {{min_selections}}.
    Maximum number of options that must be selected: {{max_selections}}.
{% elif min_selections != None %}
    Minimum number of options that must be selected: {{min_selections}}.
{% elif max_selections != None %}
    Maximum number of options that must be selected: {{max_selections}}.
{% endif %}

{# Answering Instructions #}
Please respond with valid JSON, formatted like so:
{% if include_comment %}
    {"answer": [<put comma-separated list here>], "comment": "<put explanation here>"}
{% else %}
    {"answer": [<put comma-separated list here>]}
{% endif %}

