Return your answers on one line, in a comma-separated list of your responses, with square brackets and each answer in quotes E.g., ["A", "B", "C"]
{% if include_comment %}
After the answers, you can put a comment explaining your choice on the next line.
{% endif %}